# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Pythonic Point of Sale System
#  Copyright © 2023 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - feedback control
"""

import time

import flet as ft

from .base import WuttaControl
from .keyboard import WuttaKeyboard


class WuttaFeedback(WuttaControl):

    default_font_size = 20
    default_button_height = 60

    def __init__(self, config, *args, **kwargs):
        self.on_send = kwargs.pop('on_send', None)
        self.on_cancel = kwargs.pop('on_cancel', None)

        super().__init__(config, *args, **kwargs)

    def build(self):

        self.button = ft.Container(content=ft.Text("Feedback", size=self.default_font_size,
                                                   weight=ft.FontWeight.BOLD),
                                   height=self.default_button_height,
                                   width=self.default_button_height * 3,
                                   on_click=self.initial_click,
                                   alignment=ft.alignment.center,
                                   border=ft.border.all(1, 'black'),
                                   border_radius=ft.border_radius.all(5),
                                   bgcolor='blue')

        return self.button

    def initial_click(self, e):

        self.message = ft.TextField(label="Message",
                                    multiline=True, min_lines=5,
                                    autofocus=True)

        self.dlg = ft.AlertDialog(
            modal=True,
            title=ft.Text("User Feedback"),
            content=ft.Container(
                content=ft.Column(
                    [
                        ft.Text("Questions, suggestions, comments, complaints, etc. "
                                "are welcome and may be submitted below. "),
                        ft.Divider(),
                        self.message,
                        ft.Divider(),
                        WuttaKeyboard(self.config, on_keypress=self.keypress,
                                      on_long_backspace=self.long_backspace),
                    ],
                    expand=True,
                ),
                height=800,
            ),
            actions=[
                ft.Row(
                    [
                        ft.Container(content=ft.Text("Send Message",
                                                     size=self.default_font_size,
                                                     color='black',
                                                     weight=ft.FontWeight.BOLD),
                                     height=self.default_button_height,
                                     width=self.default_button_height * 3,
                                     alignment=ft.alignment.center,
                                     bgcolor='blue',
                                     border=ft.border.all(1, 'black'),
                                     border_radius=ft.border_radius.all(5),
                                     on_click=self.send_feedback),
                        ft.Container(content=ft.Text("Cancel",
                                                     size=self.default_font_size,
                                                     weight=ft.FontWeight.BOLD),
                                     height=self.default_button_height,
                                     width=self.default_button_height * 2.5,
                                     alignment=ft.alignment.center,
                                     border=ft.border.all(1, 'black'),
                                     border_radius=ft.border_radius.all(5),
                                     on_click=self.cancel),
                    ],
                    alignment=ft.MainAxisAlignment.CENTER,
                ),
            ],
        )

        if self.mypage.dialog and self.mypage.dialog.open and self.mypage.dialog is not self.dlg:
            self.mypage.dialog.open = False
            self.mypage.update()
            # cf. https://github.com/flet-dev/flet/issues/1670
            time.sleep(0.1)

        self.mypage.dialog = self.dlg
        self.dlg.open = True
        self.mypage.update()

    def keypress(self, key):
        if key == '⏎':
            self.message.value += '\n'
        elif key == '⌫':
            self.message.value = self.message.value[:-1]
        else:
            self.message.value += key

        self.message.focus()

        # TODO: why is keypress happening with no page?
        if self.page:
            self.update()

    def long_backspace(self):
        self.message.value = self.message.value[:-10]
        self.message.focus()
        self.update()

    def cancel(self, e):
        self.dlg.open = False
        self.mypage.update()

        if self.on_cancel:
            self.on_cancel(e)

    def send_feedback(self, e):
        if self.message.value:

            self.app.send_email('pos_feedback', data={
                'user_name': self.mypage.session.get('user_display'),
                'message': self.message.value,
            })

            self.dlg.open = False
            self.show_snackbar("MESSAGE WAS SENT", bgcolor='green')
            self.mypage.update()

            if self.on_send:
                self.on_send()

        else:
            self.message.focus()
            self.mypage.update()



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Pythonic Point of Sale System
#  Copyright © 2023 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - transaction lookup control
"""

from .lookup import WuttaLookup


class WuttaTransactionLookup(WuttaLookup):

    def __init__(self, *args, **kwargs):

        # nb. this forces first query
        kwargs.setdefault('initial_search', True)

        # TODO: how to deal with 'modes'
        self.mode = kwargs.pop('mode', None)
        if not self.mode:
            raise ValueError("must specify mode")
        if self.mode != 'resume':
            raise ValueError("only 'resume' mode is supported")

        kwargs.setdefault('show_search', False)

        super().__init__(*args, **kwargs)

    def get_results_columns(self):
        return [
            "Date/Time",
            "Terminal",
            "Txn ID",
            "Cashier",
            "Customer",
            "Balance",
        ]

    def get_results(self, session, entry):
        model = self.app.model

        # TODO: how to deal with 'modes'
        assert self.mode == 'resume'
        training = bool(self.mypage.session.get('training'))
        query = session.query(model.POSBatch)\
                       .filter(model.POSBatch.status_code == model.POSBatch.STATUS_SUSPENDED)\
                       .filter(model.POSBatch.executed == None)\
                       .filter(model.POSBatch.training_mode == training)\
                       .order_by(model.POSBatch.created.desc())

        transactions = []
        for batch in query:
            # TODO: should use 'suspended' timestamp instead here?
            dt = self.app.localtime(batch.created, from_utc=True)
            transactions.append({
                'uuid': batch.uuid,
                'datetime': self.app.render_datetime(dt),
                'terminal': batch.terminal_id,
                'txnid': batch.id_str,
                'cashier': str(batch.cashier or ''),
                'customer': str(batch.customer or ''),
                'balance': self.app.render_currency(batch.get_balance()),
            })
        return transactions

    def make_result_row(self, txn):
        return [
            txn['datetime'],
            txn['terminal'],
            txn['txnid'],
            txn['cashier'],
            txn['customer'],
            txn['balance'],
        ]



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Pythonic Point of Sale System
#  Copyright © 2023 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - ten-key control
"""

import flet as ft

from .base import WuttaControl


class WuttaTenkeyMenu(WuttaControl):

    default_font_size = 40
    default_button_size = 100

    def __init__(self, *args, **kwargs):
        self.simple = kwargs.pop('simple', False)
        self.on_char = kwargs.pop('on_char', None)
        self.on_enter = kwargs.pop('on_enter', None)
        self.on_up_click = kwargs.pop('on_up_click', None)
        self.on_up_longpress = kwargs.pop('on_up_longpress', None)
        self.on_down_click = kwargs.pop('on_down_click', None)
        self.on_down_longpress = kwargs.pop('on_down_longpress', None)
        super().__init__(*args, **kwargs)

    def build(self):

        row1 = [
            self.make_tenkey_button("1"),
            self.make_tenkey_button("2"),
            self.make_tenkey_button("3"),
        ]
        if not self.simple:
            row1.extend([
                self.make_tenkey_button("-"),
            ])

        row2 = [
            self.make_tenkey_button("4"),
            self.make_tenkey_button("5"),
            self.make_tenkey_button("6"),
        ]
        if not self.simple:
            row2.extend([
                self.make_tenkey_button("↑", on_long_press=self.up_long_press),
            ])

        row3 = [
            self.make_tenkey_button("7"),
            self.make_tenkey_button("8"),
            self.make_tenkey_button("9"),
        ]
        if not self.simple:
            row3.extend([
                self.make_tenkey_button("↓", on_long_press=self.down_long_press),
            ])

        row4 = [
            self.make_tenkey_button("0"),
            # self.make_tenkey_button("00"),
            self.make_tenkey_button("."),
        ]
        if self.simple:
            row4.extend([
                self.make_tenkey_button("⏎"),
            ])
        else:
            row4.extend([
                self.make_tenkey_button("ENTER", width=self.default_button_size * 2),
            ])

        return ft.Container(
            content=ft.Column(
                [
                    ft.Row(
                        controls=row1,
                        spacing=0,
                    ),
                    ft.Row(
                        controls=row2,
                        spacing=0,
                    ),
                    ft.Row(
                        controls=row3,
                        spacing=0,
                    ),
                    ft.Row(
                        controls=row4,
                        spacing=0,
                    ),
                ],
                spacing=0,
            ),
        )

    def make_tenkey_button(
            self,
            text,
            font_size=None,
            height=None,
            width=None,
            on_click=None,
            on_long_press=None,
    ):
        if not font_size:
            font_size = self.default_font_size
        if not height:
            height = self.default_button_size
        if not width:
            width = self.default_button_size
        if not on_click:
            on_click = self.tenkey_click

        return self.make_button(text, font_size=font_size,  bgcolor='green',
                                height=height, width=width,
                                on_click=on_click,
                                on_long_press=on_long_press)

    def tenkey_click(self, e):
        value = e.control.content.value

        if value in ('ENTER', '⏎'):
            if self.on_enter:
                self.on_enter(e)

        elif value == '↑': # UP
            if self.on_up_click:
                self.on_up_click(e)

        elif value == '↓': # DOWN
            if self.on_down_click:
                self.on_down_click(e)

        else: # normal char key
            if self.on_char:
                self.on_char(value)

    def up_long_press(self, e):
        if self.on_up_longpress:
            self.on_up_longpress(e)

    def down_long_press(self, e):
        if self.on_down_longpress:
            self.on_down_longpress(e)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Pythonic Point of Sale System
#  Copyright © 2023-2024 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS commands
"""

import logging

import typer

from rattail.files import resource_path
from rattail.commands.typer import typer_callback


log = logging.getLogger(__name__)


wuttapos_typer = typer.Typer(
    callback=typer_callback,
    help="WuttaPOS (point of sale)"
)


@wuttapos_typer.command('open')
def typer_open(
        ctx: typer.Context,
):
    """
    Open the Point of Sale app
    """
    from wuttapos.app import run_app

    config = ctx.parent.rattail_config
    run_app(config)


@wuttapos_typer.command()
def serve(
        ctx: typer.Context,
):
    """
    Run the POS app as a web service
    """
    import flet as ft
    from wuttapos.app import main

    config = ctx.parent.rattail_config
    kw = {}

    host = config.get('wuttapos', 'serve.host',
                      default='0.0.0.0')
    kw['host'] = host

    port = config.getint('wuttapos', 'serve.port',
                         default=8332)
    kw['port'] = port

    # TODO: we technically "support" this, in that we do pass the
    # value on to Flet, but in practice it does not work right
    path = config.get('wuttapos', 'serve.path', default='')
    if path:
        path = path.strip('/') + '/'
        kw['name'] = path
        # kw['route_url_strategy'] = 'hash'

    log.info(f"will serve WuttaPOS on http://{host}:{port}/{path}")
    ft.app(target=main, view=None,
           assets_dir=resource_path('wuttapos:assets'),
           **kw)


@wuttapos_typer.command()
def status(
        ctx: typer.Context,
):
    """
    Show status of the POS lane
    """
    print("TODO: show status")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Pythonic Point of Sale System
#  Copyright © 2023 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS utilities
"""

import flet as ft


def get_pos_batch_handler(config):
    """
    Official way of obtaining the POS batch handler.

    Code should use this where possible to make later refactoring
    easier, should it be needed.
    """
    app = config.get_app()
    return app.get_batch_handler('pos', default='rattail.batch.pos:POSBatchHandler')


def make_button(text, font_size=24, font_bold=True, font_weight=None, **kwargs):
    """
    Generic function for making a Container button.
    """
    if not font_weight and font_bold:
        font_weight = ft.FontWeight.BOLD
    text = ft.Text(text, size=font_size, weight=font_weight,
                   text_align=ft.TextAlign.CENTER)

    kwargs.setdefault('alignment', ft.alignment.center)
    kwargs.setdefault('border', ft.border.all(1, 'black'))
    kwargs.setdefault('border_radius', ft.border_radius.all(5))
    return ft.Container(content=text, **kwargs)


def show_snackbar(page, text, bgcolor='yellow'):
    page.snack_bar = ft.SnackBar(ft.Text(text, color='black',
                                         size=40,
                                         weight=ft.FontWeight.BOLD),
                                 bgcolor=bgcolor,
                                 duration=1500)
    page.snack_bar.open = True



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Pythonic Point of Sale System
#  Copyright © 2023 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - txn item control
"""

import flet as ft

from .base import WuttaControl


class WuttaTxnItem(WuttaControl):
    """
    Control for displaying a transaction line item within main POS
    items list.
    """
    font_size = 24

    def __init__(self, config, row, *args, **kwargs):
        super().__init__(config, *args, **kwargs)
        self.row = row

    def build(self):

        self.major_style = ft.TextStyle(size=self.font_size,
                                        weight=ft.FontWeight.BOLD)

        self.minor_style = ft.TextStyle(size=int(self.font_size * 0.8),
                                        italic=True)

        if self.row.row_type in (self.enum.POS_ROW_TYPE_SELL,
                                 self.enum.POS_ROW_TYPE_OPEN_RING):
            return self.build_item_sell()

        elif self.row.row_type in (self.enum.POS_ROW_TYPE_TENDER,
                                   self.enum.POS_ROW_TYPE_CHANGE_BACK):
            return self.build_item_tender()

    def build_item_sell(self):

        self.quantity = ft.TextSpan(style=self.minor_style)
        self.txn_price = ft.TextSpan(style=self.minor_style)

        self.sales_total_style = ft.TextStyle(size=self.font_size,
                                              weight=ft.FontWeight.BOLD)

        self.sales_total = ft.TextSpan(style=self.sales_total_style)

        self.fs_flag = ft.TextSpan(style=self.minor_style)
        self.tax_flag = ft.TextSpan(style=self.minor_style)

        # set initial text display values
        self.refresh(update=False)

        return ft.Row(
            [
                ft.Text(
                    spans=[
                        ft.TextSpan(f"{self.row.description}",
                                    style=self.major_style),
                        ft.TextSpan("× ", style=self.minor_style),
                        self.quantity,
                        ft.TextSpan(" @ ", style=self.minor_style),
                        self.txn_price,
                    ],
                ),
                ft.Text(
                    spans=[
                        self.fs_flag,
                        self.tax_flag,
                        self.sales_total,
                    ],
                ),

            ],
            alignment=ft.MainAxisAlignment.SPACE_BETWEEN,
        )

    def build_item_tender(self):
        return ft.Row(
            [
                ft.Text(
                    spans=[
                        ft.TextSpan(f"{self.row.description}",
                                    style=self.major_style),
                    ],
                ),
                ft.Text(
                    spans=[
                        ft.TextSpan(self.app.render_currency(self.row.tender_total),
                                    style=self.major_style),
                    ],
                ),

            ],
            alignment=ft.MainAxisAlignment.SPACE_BETWEEN,
        )

    def refresh(self, update=True):

        if self.row.void:
            self.major_style.color = None
            self.major_style.decoration = ft.TextDecoration.LINE_THROUGH
            self.major_style.weight = None
            self.minor_style.color = None
            self.minor_style.decoration = ft.TextDecoration.LINE_THROUGH
        else:
            self.major_style.color = None
            self.major_style.decoration = None
            self.major_style.weight = ft.FontWeight.BOLD
            self.minor_style.color = None
            self.minor_style.decoration = None

        if self.row.row_type in (self.enum.POS_ROW_TYPE_SELL,
                                 self.enum.POS_ROW_TYPE_OPEN_RING):
            self.quantity.text = self.app.render_quantity(self.row.quantity)
            self.txn_price.text = self.app.render_currency(self.row.txn_price)
            self.sales_total.text = self.app.render_currency(self.row.sales_total)
            self.fs_flag.text = "FS   " if self.row.foodstamp_eligible else ""
            self.tax_flag.text = f"T{self.row.tax_code}   " if self.row.tax_code else ""

            if self.row.void:
                self.sales_total_style.color = None
                self.sales_total_style.decoration = ft.TextDecoration.LINE_THROUGH
                self.sales_total_style.weight = None
            else:
                if (self.row.row_type == self.enum.POS_ROW_TYPE_SELL
                    and self.row.txn_price_adjusted):
                    self.sales_total_style.color = 'orange'
                elif (self.row.row_type == self.enum.POS_ROW_TYPE_SELL
                      and self.row.cur_price and self.row.cur_price < self.row.reg_price):
                    self.sales_total_style.color = 'green'
                else:
                    self.sales_total_style.color = None
                self.sales_total_style.decoration = None
                self.sales_total_style.weight = ft.FontWeight.BOLD

        if update:
            self.update()



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Pythonic Point of Sale System
#  Copyright © 2023 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - department lookup control
"""

from .lookup import WuttaLookup


class WuttaDepartmentLookup(WuttaLookup):

    def __init__(self, *args, **kwargs):

        # nb. this forces first query
        kwargs.setdefault('initial_search', '')
        kwargs.setdefault('allow_empty_query', True)

        super().__init__(*args, **kwargs)

    def get_results_columns(self):
        return [
            "Number",
            "Name",
        ]

    def get_results(self, session, entry):
        model = self.app.model
        query = session.query(model.Department)\
                       .order_by(model.Department.number)

        if entry:
            query = query.filter(model.Department.name.ilike(f'%{entry}%'))

        departments = []
        for dept in query:
            departments.append({
                'uuid': dept.uuid,
                'number': str(dept.number),
                'name': dept.name,
            })
        return departments

    def make_result_row(self, dept):
        return [
            dept['number'],
            dept['name'],
        ]



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Pythonic Point of Sale System
#  Copyright © 2023 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - login form control
"""

import flet as ft

from .base import WuttaControl
from .keyboard import WuttaKeyboard
from .tenkey import WuttaTenkeyMenu


class WuttaLoginForm(WuttaControl):

    def __init__(self, config, *args, **kwargs):

        # permission to be checked for login to succeed
        self.perm_required = kwargs.pop('perm_required', 'pos.ring_sales')

        # may or may not show the username field
        # nb. must set this before normal __init__
        if 'show_username' in kwargs:
            self.show_username = kwargs.pop('show_username')
        else:
            self.show_username = config.getbool('wuttapos', 'login.show_username',
                                                default=True)

        # may or may not show 10-key menu instead of full keyboard
        if 'use_tenkey' in kwargs:
            self.use_tenkey = kwargs.pop('use_tenkey')
        else:
            self.use_tenkey = config.getbool('wuttapos', 'login.use_tenkey',
                                             default=False)

        self.on_login_failure = kwargs.pop('on_login_failure', None)
        self.on_authz_failure = kwargs.pop('on_authz_failure', None)
        self.on_login_success = kwargs.pop('on_login_success', None)

        super().__init__(config, *args, **kwargs)

        # track which login input has focus
        self.focused = None

    def build(self):

        login_form = self.build_login_form()

        if self.use_tenkey:
            controls = [
                ft.Row(
                    [
                        login_form,
                        ft.VerticalDivider(),
                        WuttaTenkeyMenu(self.config, simple=True,
                                        on_char=self.tenkey_char,
                                        on_enter=self.tenkey_enter),
                    ],
                    alignment=ft.MainAxisAlignment.CENTER,
                ),
            ]

        else: # full keyboard
            controls = [
                login_form,
                ft.Row(),
                ft.Row(),
                ft.Row(),
                WuttaKeyboard(self.config, on_keypress=self.keyboard_keypress,
                              on_long_backspace=self.keyboard_long_backspace),
            ]

        return ft.Column(controls=controls,
                         expand=True,
                         alignment=ft.MainAxisAlignment.CENTER)

    def build_login_form(self):
        form_fields = []

        self.password = ft.TextField(label="Password", width=200, password=True,
                                     on_submit=self.password_submit,
                                     on_focus=self.password_focus,
                                     autofocus=not self.show_username)
        self.focused = self.password

        if self.show_username:
            self.username = ft.TextField(label="Login", width=200,
                                         on_submit=self.username_submit,
                                         on_focus=self.username_focus,
                                         autofocus=True)
            form_fields.append(self.username)
            self.focused = self.username

        form_fields.append(self.password)

        login_button = self.make_button("Login",
                                        height=60,
                                        width=60 * 2.5,
                                        bgcolor='blue',
                                        on_click=self.attempt_login)

        reset_button = self.make_button("Clear",
                                        height=60,
                                        width=60 * 2.5,
                                        on_click=self.clear_login)

        if self.use_tenkey:
            form_fields.extend([
                ft.Row(),
                ft.Row(),
                ft.Row(
                    [
                        reset_button,
                        login_button,
                    ],
                ),
            ])
            return ft.Column(
                controls=form_fields,
                horizontal_alignment=ft.CrossAxisAlignment.CENTER,
            )

        else: # full keyboard
            form_fields.extend([
                login_button,
                reset_button,
            ])
            return ft.Row(
                controls=form_fields,
                alignment=ft.MainAxisAlignment.CENTER,
            )

    def keyboard_keypress(self, key):
        assert self.focused

        if key == '⏎': # ENTER

            # attempt to submit the login form..
            if self.show_username and self.focused is self.username:
                self.username_submit()
            else:
                if self.password_submit():

                    # here the login has totally worked, which means
                    # this form has fulfilled its purpose.  hence must
                    # exit early to avoid update() in case we are to
                    # be redirected etc.  otherwise may get errors
                    # trying to update controls which have already
                    # been dropped from the page..
                    return

        elif key == '⌫':
            self.focused.value = self.focused.value[:-1]

        else:
            self.focused.value += key

        self.focused.focus()
        self.update()

    def keyboard_long_backspace(self):
        assert self.focused
        self.focused.value = ''
        self.focused.focus()
        self.update()

    def tenkey_char(self, key):
        if key == '@':
            return

        self.focused.value = f"{self.focused.value or ''}{key}"
        self.update()

    def tenkey_enter(self, e):
        if self.show_username and self.focused is self.username:
            self.username_submit(e)
            self.update()
        else:
            if not self.password_submit(e):
                self.update()

    def username_focus(self, e):
        self.focused = self.username

    def username_submit(self, e=None):
        if self.username.value:
            self.password.focus()
        else:
            self.username.focus()

    def password_focus(self, e):
        self.focused = self.password

    def password_submit(self, e=None):
        if self.password.value:
            return self.attempt_login(e)
        else:
            self.password.focus()
            return False

    def attempt_login(self, e=None):
        if self.show_username and not self.username.value:
            self.username.focus()
            return False
        if not self.password.value:
            self.password.focus()
            return False

        session = self.app.make_session()
        auth = self.app.get_auth_handler()
        user = auth.authenticate_user(session,
                                      self.username.value if self.show_username else None,
                                      self.password.value)
        user_display = str(user) if user else None
        has_perm = auth.has_permission(session, user, self.perm_required) if user else False
        session.close()

        if user:

            if has_perm:
                if self.on_login_success:
                    self.on_login_success(user, user_display)
                return True

            else:
                if self.on_authz_failure:
                    self.on_authz_failure(user, user_display)
                self.clear_login()

        else:
            if self.on_login_failure:
                self.on_login_failure(e)
            self.clear_login()

        return False

    def clear_login(self, e=None):
        if self.show_username:
            self.username.value = ""
        self.password.value = ""
        if self.show_username:
            self.username.focus()
        else:
            self.password.focus()
        self.update()



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Pythonic Point of Sale System
#  Copyright © 2023 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - custom controls (base class)
"""

import flet as ft

from wuttapos.util import make_button, show_snackbar


class WuttaControl(ft.UserControl):

    def __init__(self, config, page=None, *args, **kwargs):
        self.on_reset = kwargs.pop('on_reset', None)

        super().__init__(*args, **kwargs)

        self.config = config
        self.app = config.get_app()
        self.enum = self.app.enum

        # TODO: why must we save this aside from self.page ?
        # but sometimes self.page gets set to None, so we must..
        self.mypage = page

    def informed_refresh(self, **kwargs):
        pass

    def make_button(self, *args, **kwargs):
        return make_button(*args, **kwargs)

    def reset(self, e=None):
        if self.on_reset:
            self.on_reset(e=e)

    def show_snackbar(self, text, bgcolor='yellow'):
        show_snackbar(self.mypage, text, bgcolor=bgcolor)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Pythonic Point of Sale System
#  Copyright © 2023 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - keyboard control
"""

import flet as ft

from .base import WuttaControl


class WuttaKeyboard(WuttaControl):

    default_font_size = 20
    default_button_size = 80

    def __init__(self, *args, **kwargs):
        self.on_keypress = kwargs.pop('on_keypress', None)
        self.on_long_backspace = kwargs.pop('on_long_backspace', None)

        super().__init__(*args, **kwargs)

        self.caps_lock = False
        self.caps_map = dict([(k, k.upper())
                              for k in 'abcdefghijklmnopqrstuvwxyz'])

        self.shift = False
        self.shift_map = {
            '`': '~',
            '1': '!',
            '2': '@',
            '3': '#',
            '4': '$',
            '5': '%',
            '6': '^',
            '7': '&',
            '8': '*',
            '9': '(',
            '0': ')',
            '-': '_',
            '=': '+',
            '[': '{',
            ']': '}',
            '\\': '|',
            ';': ':',
            "'": '"',
            ',': '<',
            '.': '>',
            '/': '?',
        }

    def update_caps_lock(self, caps_lock):
        self.caps_lock = caps_lock

        if self.caps_lock:
            self.caps_key.bgcolor = 'blue'
        else:
            self.caps_key.bgcolor = None

        for key, button in self.keys.items():
            if key in self.caps_map:
                if self.caps_lock or self.shift:
                    button.content.value = self.caps_map[key]
                else:
                    button.content.value = key

        self.update()

    def update_shift(self, shift):
        self.shift = shift

        if self.shift:
            self.shift_key.bgcolor = 'blue'
        else:
            self.shift_key.bgcolor = None

        for key, button in self.keys.items():
            if key in self.caps_map:
                if self.shift or self.caps_lock:
                    button.content.value = self.caps_map[key]
                else:
                    button.content.value = key
            elif key in self.shift_map:
                if self.shift:
                    button.content.value = self.shift_map[key]
                else:
                    button.content.value = key

        self.update()

    def simple_keypress(self, e):

        # maybe inform parent
        if self.on_keypress:
            key = e.control.content.value

            # avoid callback for certain keys
            if key not in ('CAPS', 'SHIFT'):

                # translate certain keys
                if key == 'SPACE':
                    key = ' '

                # let 'em know
                self.on_keypress(key)

        # turn off shift key if set
        if self.shift:
            self.update_shift(False)

    def long_backspace(self, e):
        if self.on_long_backspace:
            self.on_long_backspace()

    def build(self):
        self.keys = {}

        def make_key(key, data=None,
                     on_click=self.simple_keypress,
                     on_long_press=None,
                     width=self.default_button_size,
                     bgcolor=None):
            button = ft.Container(content=ft.Text(key, size=self.default_font_size,
                                                  weight=ft.FontWeight.BOLD),
                                  data=data or key,
                                  height=self.default_button_size,
                                  width=width,
                                  on_click=on_click,
                                  on_long_press=on_long_press,
                                  alignment=ft.alignment.center,
                                  border=ft.border.all(1, 'black'),
                                  border_radius=ft.border_radius.all(5),
                                  bgcolor=bgcolor,
                                  )
            self.keys[key] = button
            return button

        def caps_click(e):
            self.update_caps_lock(not self.caps_lock)

        self.caps_key = make_key('CAPS', on_click=caps_click)
        if self.caps_lock:
            self.caps_key.bgcolor = 'blue'

        def shift_click(e):
            self.update_shift(not self.shift)

        self.shift_key = make_key('SHIFT', on_click=shift_click)
        if self.shift:
            self.shift_key.bgcolor = 'blue'

        rows = [
            [make_key(k) for k in "`1234567890-="] + [make_key('⌫', bgcolor='yellow',
                                                               on_long_press=self.long_backspace)],
            [make_key(k) for k in "qwertyuiop[]\\"],
            [self.caps_key] + [make_key(k) for k in "asdfghjkl;'"] + [make_key('⏎', bgcolor='blue')],
            [self.shift_key] + [make_key(k) for k in "zxcvbnm,./"],
            [make_key('SPACE', width=self.default_button_size * 5)],
        ]

        rows = [ft.Row(controls, alignment=ft.MainAxisAlignment.CENTER)
                for controls in rows]

        return ft.Container(
            content=ft.Column(
                rows,
            ),
        )



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Pythonic Point of Sale System
#  Copyright © 2023 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - timestamp control
"""

import threading
import time

import flet as ft

from .base import WuttaControl


class WuttaTimestamp(WuttaControl):

    def __init__(self, *args, **kwargs):
        self.weight = kwargs.pop('weight', None)
        self.size = kwargs.pop('size', None)
        super().__init__(*args, **kwargs)

    def build(self):
        text = self.render_time(self.app.localtime())
        self.display = ft.Text(text, weight=self.weight, size=self.size)

        thread = threading.Thread(target=self.update_display)
        thread.start()

        return self.display

    def render_time(self, value):
        return value.strftime('%a %d %b %Y %I:%M:%S %p')

    def update_display(self):
        while True:
            if self.page:
                self.display.value = self.render_time(self.app.localtime())
                self.update()
            time.sleep(0.5)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaPOS -- Pythonic Point of Sale System
#  Copyright © 2023 Lance Edgar
#
#  This file is part of WuttaPOS.
#
#  WuttaPOS is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
WuttaPOS - header control
"""

import rattail

import flet as ft

import wuttapos
from .base import WuttaControl
from .timestamp import WuttaTimestamp
from .feedback import WuttaFeedback


class WuttaHeader(WuttaControl):

    def __init__(self, *args, **kwargs):
        self.term_display = kwargs.pop('term_display', '??')
        super().__init__(*args, **kwargs)

    def build(self):
        self.txn_display = ft.Text("Txn: N", weight=ft.FontWeight.BOLD, size=20)
        self.cust_display = ft.Text("Cust: N", weight=ft.FontWeight.BOLD, size=20)

        self.training_mode = ft.Text(size=40, weight=ft.FontWeight.BOLD)

        self.user_display = ft.Text("User: N", weight=ft.FontWeight.BOLD, size=20)
        self.logout_button = ft.OutlinedButton("Logout", on_click=self.logout_click, visible=False)
        self.logout_divider = ft.VerticalDivider(visible=False)
        self.title_button = ft.FilledButton(self.app.get_title(), on_click=self.title_click)

        return ft.Stack(
            controls=[
                ft.Container(
                    content=ft.Row(
                        [
                            ft.Container(
                                content=self.training_mode,
                                bgcolor='yellow',
                            ),
                        ],
                        alignment=ft.MainAxisAlignment.CENTER,
                    ),
                ),
                ft.Row(
                    [
                        ft.Row(
                            [
                                self.txn_display,
                                ft.VerticalDivider(),
                                self.cust_display,
                                ft.VerticalDivider(),
                                WuttaTimestamp(self.config, weight=ft.FontWeight.BOLD, size=20),
                            ],
                        ),
                        ft.Row(
                            [
                                self.user_display,
                                ft.VerticalDivider(),
                                self.logout_button,
                                self.logout_divider,
                                ft.Text(f"Term: {self.term_display}", weight=ft.FontWeight.BOLD, size=20),
                                ft.VerticalDivider(),
                                self.title_button,
                            ],
                        ),
                    ],
                    alignment=ft.MainAxisAlignment.SPACE_BETWEEN,
                ),

            ],
        )

    def did_mount(self):
        self.informed_refresh()

    def informed_refresh(self):
        self.update_txn_display()
        self.update_cust_display()
        self.update_user_display()
        self.update_training_display()
        self.update()

    def update_txn_display(self):
        txn_display = None
        if self.page:
            txn_display = self.page.session.get('txn_display')
        self.txn_display.value = f"Txn: {txn_display or 'N'}"

    def update_cust_display(self):
        cust_display = None
        if self.page:
            cust_display = self.page.session.get('cust_display')
        self.cust_display.value = f"Cust: {cust_display or 'N'}"

    def update_training_display(self):
        if self.page.session.get('training'):
            self.training_mode.value = "  TRAINING MODE  "
        else:
            self.training_mode.value = ""

    def update_user_display(self):
        user_display = None
        if self.page:
            user_display = self.page.session.get('user_display')
        self.user_display.value = f"User: {user_display or 'N'}"

        if self.page and self.page.session.get('user_uuid'):
            self.logout_button.visible = True
            self.logout_divider.visible = True

    def logout_click(self, e):

        # TODO: hacky but works for now
        if not self.config.production():
            self.page.client_storage.set('user_uuid', '')

        self.page.session.clear()
        self.page.go('/login')

    def title_click(self, e):
        title = self.app.get_title()

        year = self.app.today().year
        if year > 2023:
            year_range = f'2023 - {year}'
        else:
            year_range = year

        license = f"""\
WuttaPOS -- Pythonic Point of Sale System
Copyright © {year_range} Lance Edgar

WuttaPOS is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

WuttaPOS is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
WuttaPOS.  If not, see <http://www.gnu.org/licenses/>.
"""

        buttons = []

        user_uuid = self.page.session.get('user_uuid')
        if user_uuid:

            session = self.app.make_session()
            model = self.app.model
            user = session.get(model.User, user_uuid)
            auth = self.app.get_auth_handler()
            has_perm = auth.has_permission(session, user, 'pos.test_error')
            session.close()
            if has_perm:
                test_error = self.make_button("TEST ERROR", font_size=24,
                                              height=60,
                                              width=60 * 3,
                                              bgcolor='red',
                                              on_click=self.test_error_click)
                buttons.append(test_error)

            feedback = WuttaFeedback(self.config, page=self.page,
                                     on_send=self.reset, on_cancel=self.reset)
            buttons.append(feedback)

        self.dlg = ft.AlertDialog(
            title=ft.Text(title),
            content=ft.Container(
                content=ft.Column(
                    [
                        ft.Divider(),
                        ft.Text(f"{title} v{wuttapos.__version__}"),
                        ft.Text(f"rattail v{rattail.__version__}"),
                        ft.Divider(),
                        ft.Text(license),
                        ft.Container(
                            content=ft.Row(
                                controls=buttons,
                                alignment=ft.MainAxisAlignment.CENTER,
                            ),
                            alignment=ft.alignment.center,
                            expand=True,
                        ),
                    ],
                    expand=True,
                ),
                height=600,
            ),
            actions=[
                ft.Row(
                    [
                        ft.Container(content=ft.Text("Close", size=20,
                                                     weight=ft.FontWeight.BOLD),
                                     height=60,
                                     width=60 * 2.5,
                                     alignment=ft.alignment.center,
                                     border=ft.border.all(1, 'black'),
                                     border_radius=ft.border_radius.all(5),
                                     on_click=self.close_dlg),
                    ],
                    alignment=ft.MainAxisAlignment.CENTER,
                ),
            ],
        )

        self.page.dialog = self.dlg
        self.dlg.open = True
        self.page.update()

    def test_error_click(self, e):

        # first get the dialog out of the way
        self.dlg.open = False
        self.reset()
        self.page.update()

        raise RuntimeError("FAKE ERROR - to test error handling")

    def close_dlg(self, e):
        self.dlg.open = False
        self.reset()
        self.page.update()
